@@ -44,6 +44,65 @@ static bool isKeyFocusReason(Qt::FocusReason reason)
     return reason == Qt::TabFocusReason || reason == Qt::BacktabFocusReason || reason == Qt::ShortcutFocusReason;
 }
 
+// There are lots of events on QApp / QWindow
+// This filters events on those objects to only events KQuickStyleItem is interested in which can then be monitored.
+class KQuickStyleItemEventListener : public QObject
+{
+    Q_OBJECT
+public:
+    static QSharedPointer<KQuickStyleItemEventListener> listenerForWindow(QWindow *window);
+    bool eventFilter(QObject *watched, QEvent *event) override;
+
+private:
+    KQuickStyleItemEventListener(QWindow *);
+    static QHash<QWindow *, QWeakPointer<KQuickStyleItemEventListener>> s_listeners;
+};
+
+QHash<QWindow *, QWeakPointer<KQuickStyleItemEventListener>> KQuickStyleItemEventListener::s_listeners;
+
+KQuickStyleItemEventListener::KQuickStyleItemEventListener(QWindow *window)
+    : QObject()
+{
+    window->installEventFilter(this);
+    qApp->installEventFilter(this);
+}
+
+QSharedPointer<KQuickStyleItemEventListener> KQuickStyleItemEventListener::listenerForWindow(QWindow *window)
+{
+    Q_ASSERT(window);
+    if (auto listener = s_listeners.value(window).lock()) {
+        return listener;
+    }
+
+    auto listener = QSharedPointer<KQuickStyleItemEventListener>(
+        new KQuickStyleItemEventListener(window),
+        [window](KQuickStyleItemEventListener* ptr) {
+            delete ptr;
+            s_listeners.remove(window);
+        }
+        );
+    s_listeners.insert(window, listener.toWeakRef());
+    return listener;
+}
+
+bool KQuickStyleItemEventListener::eventFilter(QObject *watched, QEvent *event)
+{
+    if (watched == qGuiApp) {
+        if (event->type() == QEvent::ApplicationFontChange) {
+            QCoreApplication::sendEvent(this, event);
+        }
+    }
+    if (qobject_cast<QWindow *>(watched)) {
+        if (event->type() == QEvent::KeyPress || event->type() == QEvent::KeyRelease) {
+            QKeyEvent *ke = static_cast<QKeyEvent *>(event);
+            if (ke->key() == Qt::Key_Alt) {
+                QCoreApplication::sendEvent(this, event);
+            }
+        }
+    }
+    return false;
+}
+
 KQuickStyleItem::KQuickStyleItem(QQuickItem *parent)
     : QQuickItem(parent)
     , m_styleoption(nullptr)
@@ -119,8 +178,6 @@ KQuickStyleItem::KQuickStyleItem(QQuickItem *parent)
     setFlag(QQuickItem::ItemHasContents, true);
     setSmooth(false);
 
-    qGuiApp->installEventFilter(this);
-
     Kirigami::Platform::TabletModeWatcher::self()->addWatcher(this);
 }
 
@@ -172,6 +229,9 @@ void KQuickStyleItem::initStyleOption()
 
         connect(m_theme, &Kirigami::Platform::PlatformTheme::colorsChanged, this, [this]() {
             // we need to reset the palette event if Qt::AA_SetPalette attribute has been set
+            if (!m_styleoption) {
+                return;
+            }
             m_styleoption->palette = m_theme->palette();
             polish();
         });
@@ -1966,17 +2026,10 @@ void KQuickStyleItem::setControl(QQuickItem *control)
         m_control->installEventFilter(this);
 
         if (m_control->window()) {
-            m_window = m_control->window();
-            m_window->installEventFilter(this);
+            updateWindow(m_control->window());
         }
         connect(m_control, &QQuickItem::windowChanged, this, [this](QQuickWindow *window) {
-            if (m_window) {
-                m_window->removeEventFilter(this);
-            }
-            m_window = window;
-            if (m_window) {
-                m_window->installEventFilter(this);
-            }
+            updateWindow(window);
         });
     }
 
@@ -2055,16 +2108,18 @@ bool KQuickStyleItem::eventFilter(QObject *watched, QEvent *event)
                 return true;
             }
         }
-    } else if (watched == m_window.data()) {
-        if (event->type() == QEvent::KeyPress || event->type() == QEvent::KeyRelease) {
-            QKeyEvent *ke = static_cast<QKeyEvent *>(event);
-            if (ke->key() == Qt::Key_Alt) {
+    }
+    if (watched == m_eventListener) {
+        switch (event->type()) {
+            case QEvent::ApplicationFontChange:
+                QMetaObject::invokeMethod(this, &KQuickStyleItem::updateSizeHint, Qt::QueuedConnection);
+                break;
+            case QEvent::KeyPress:
+            case QEvent::KeyRelease:
                 polish();
-            }
-        }
-    } else if (watched == qGuiApp) {
-        if (event->type() == QEvent::ApplicationFontChange) {
-            QMetaObject::invokeMethod(this, &KQuickStyleItem::updateSizeHint, Qt::QueuedConnection);
+                break;
+            default:
+                Q_ASSERT(false); // other events should have been filtered out
         }
     }
 
@@ -2094,6 +2149,19 @@ void KQuickStyleItem::geometryChange(const QRectF &newGeometry, const QRectF &ol
     }
 }
 
+void KQuickStyleItem::updateWindow(QWindow *window)
+{
+    if (m_eventListener) {
+        m_eventListener->removeEventFilter(this);
+    }
+    if (window) {
+        m_eventListener = KQuickStyleItemEventListener::listenerForWindow(window);
+        m_eventListener->installEventFilter(this);
+    } else {
+        m_eventListener.reset();
+    }
+}
+
 void KQuickStyleItem::styleChanged()
 {
     // It should be safe to use qApp->style() unguarded here, because the signal
@@ -2158,5 +2226,6 @@ QPixmap QQuickTableRowImageProvider1::requestPixmap(const QString &id, QSize *si
     return pixmap;
 }
 
+#include "kquickstyleitem.moc"
 #include "moc_kquickpadding_p.cpp"
 #include "moc_kquickstyleitem_p.cpp"
